module Math
    Random = Random
end

module GA
    module Random
        @@random = Math::Random.new
        
        def self.set_seed(i)
            @@random = Math::Random.new i
        end
        
        def self.random
            @@random
        end
        
        def self.shuffle(array)
            GA.contract("Shuffle needs an array") { array.is_a?(Array) }
            
            return array.shuffle(random: @@random)
        end
        
        def self.rand(*args)
            @@random.rand *args
        end
        
        def self.rand_int(*args)
            self.rand(*args).to_i
        end
        
        def self.with_probability(probability)
            GA.contract("The probability must be in the range [0, 1] (#{probability} given)") { probability.between?(0, 1) }
            
            return self.rand <= probability
        end
        
        def self.random_sign
            value = self.rand(2)
            
            return value == 0 ? -1 : 1
        end
    end
    
    module Logger
        LEVEL_DEBUG     = 0
        LEVEL_INFO      = 1
        LEVEL_WARNING   = 2
        LEVEL_SEVERE    = 3
        
        @@level = 0
        def self.debug(*data, &block)
            return if @@level > LEVEL_DEBUG
            self.log "DEBUG", *data, &block
        end
        
        def self.info(*data, &block)
            return if @@level > LEVEL_INFO
            self.log "INFO", *data, &block
        end
        
        def self.warning(*data, &block)
            return if @@level > LEVEL_WARNING
            self.log "WARNING", *data, &block
        end
        
        def self.severe(*data, &block)
            return if @@level > LEVEL_SEVERE
            self.log "SEVERE", *data, &block
        end
        
        def self.level=(level)
            GA.contract("The logger level must be between 0 and 3 (#{level} given)") { level.between?(0, 3) }
            @@level = level
        end
        
        private
        def self.log(type, *data, &block)
            if block_given?
                d = block.call
                puts "#{type}: #{d}"
            else
                data.each do |d|
                    puts "#{type}: #{d}"
                end
            end
        end
    end
    
    def self.profile(profile)
        case profile
        when :development
            @@assumptions_checking = :hard
            GA::Logger.level = GA::Logger::LEVEL_DEBUG
            
        when :testing
            @@assumptions_checking = :hard
            GA::Logger.level = GA::Logger::LEVEL_INFO
            
        when :production
            @@assumptions_checking = :soft
            GA::Logger.level = GA::Logger::LEVEL_WARNING
            
        when :stable_production
            @@assumptions_checking = false
            GA::Logger.level = GA::Logger::LEVEL_SEVERE
            
        else
            raise "Illegal profile #{profile}"
        end
    end
    
    class ContractException < StandardError; end
    class PostConditionException < StandardError; end
    def self.contract(message = "Contract violated", condition = false, &block)
        return unless @@assumptions_checking
        
        raise TypeError, "no implicit conversion of #{message.class} into String" unless message.is_a? String
        
        any_problem = false
        if block_given?
            any_problem = true unless yield
        else
            any_problem = true unless condition
        end
        
        if any_problem
            case @@assumptions_checking
            when :hard
                raise ContractException, message
            when :soft
                GA::Logger.severe(message)
            end
        end
    end
    
    def self.postcondition(message = "Postcondition violated", condition = false, &block)
        return unless @@assumptions_checking
        
        raise TypeError, "no implicit conversion of #{message.class} into String" unless message.is_a? String
                
        any_problem = false
        if block_given?
            any_problem = true unless yield
        else
            any_problem = true unless condition
        end
        
        if any_problem
            case @@assumptions_checking
            when :hard
                raise ContractException, message
            when :soft
                GA::Logger.severe(message)
            end
        end
    end
    
    GA.profile :testing
end

def gadebug(*data)
    GA::Logger.debug(*data)
end

def gainfo(*data)
    GA::Logger.info(*data)
end

def gawarn(*data)
    GA::Logger.warning(*data)
end

def gasevere(*data)
    GA::Logger.severe(*data)
end

require_relative "genetic_algorithm"
require_relative "operators"
require_relative "solution"
